# coding: utf-8
import datetime

from django.db import IntegrityError
from django.db import models
from django.db import router
from django.db import transaction
from django.db.models.signals import pre_delete
from educommon.django.db.models import ReadOnlyMixin
from educommon.django.db.utils import model_modifier_metaclass
from m3 import ApplicationLogicException
from m3.db import BaseObjectModel
from objectpack.exceptions import ValidationError
from registry.info_registry import BaseInfoModel

from .exceptions import ItemAlreadyExistError
from .misc import is_duplicate_integrity_error


def safe_delete(obj):
    """Удаление сначала связи с другими моделями."""
    from django.db.models.deletion import Collector
    from m3 import RelatedError

    using = router.db_for_write(obj.__class__, instance=obj)
    pre_delete.send(sender=obj.__class__, instance=obj)
    collector = Collector(using=using)
    collector.collect([obj])

    # в коллекторе есть ссылка на самого себя
    collector.data[obj.__class__].remove(obj)
    if len(collector.data[obj.__class__]) == 0:
        del collector.data[obj.__class__]

    if collector.data:
        s = "У '%s' найдены зависимости (%s): " % (
            obj, len(collector.data))
        for key, value in list(collector.data.items()):
            s += "<br />" + key._meta.verbose_name.title()
            s += " (" + str(len(value)) + ")"
        s += "<br />" + (
            'Для удаления выбранных данных необходимо '
            'предварительно удалить '
            'их зависимости в системе.'
        )
        raise RelatedError(s)
    return True


class BaseModel(BaseObjectModel):
    """Базовый класс для всех моделей объектов, подлежащих репликации."""

    # момент создания записи
    created = models.DateTimeField(
        auto_now_add=True,
        verbose_name='Дата создания'
    )
    # момент изменения записи
    modified = models.DateTimeField(
        auto_now=True,
        verbose_name='Дата изменения'
    )

    report_constructor_params = {
        'except': ('created', 'modified')
    }

    def __repr__(self):
        """Строгое "официальное" отображение объекта класса.

        This should look like a valid Python expression.

        Could be used to recreate an object with the same value
        (given an appropriate environment).

        Т.е., eval(repr(self)) инстанцирует объект.

        """

        return '{cls_name}({params})'.format(
            cls_name=type(self).__qualname__,
            params=', '.join(
                '{0}={1}'.format(f.attname, f.value_to_string(self))
                for f in self._meta.fields
            )
        )

    def save(self, *args, **kwargs):
        for f in self.__class__._meta.fields:
            if isinstance(f, models.BooleanField):
                if not f.null and getattr(self, f.name) is None:
                    setattr(self, f.name, f.default)
            if isinstance(f, (models.TextField, models.CharField)):
                try:
                    val = getattr(self, f.attname)
                except AttributeError:
                    continue
                if isinstance(val, str):
                    val = val.strip()
                    if f.max_length and len(val) > f.max_length:
                        raise ValidationError(
                            'Превышена длина поля "%s"!' % f.attname)
                val = val or None
                setattr(self, f.attname, val)
            if isinstance(f, models.IntegerField):
                try:
                    val = getattr(self, f.attname)
                except AttributeError:
                    continue
                if val and not (-2147483648 <= int(val) <= 2147483647):
                    raise ApplicationLogicException(
                        'Превышено значение целого числа.')
            if isinstance(f, models.FileField):
                try:
                    val = getattr(self, f.attname)
                except AttributeError:
                    continue
                if not val.name:
                    continue
                if f.max_length and len(val.name) > f.max_length:
                    raise ValidationError(
                        'Превышена длина поля "%s"!' % f.attname)

            if isinstance(f, (models.DateField, models.DateTimeField)):
                if isinstance(self, BaseInfoModel):
                    continue
                try:
                    val = getattr(self, f.attname)
                except AttributeError:
                    continue
                if not isinstance(val, (datetime.datetime, datetime.date)):
                    continue

                if val.year < 1900:
                    raise ApplicationLogicException(
                        'Дата должна быть больше 1 января 1900 года!')

                setattr(self, f.attname, val)

        super(BaseModel, self).save(*args, **kwargs)

    @transaction.atomic
    def safe_delete(self):
        safe_delete(self)
        return super(BaseModel, self).safe_delete()

    @classmethod
    def get_object_relations(cls, obj_id):
        """
        Функция для определения моделей и идентификаторов,
        объектов зависимых от экземпляра данного класса.
        Экземпляр класса задается через cls и obj_id

        :param cls: Класс экземпляра, от которого будем искать зависимых
        :param obj_id: ID экземпляра от которого будем искать зависимых
        :return: Кортеж кортежей вида:
            (
                (ModelClass_1, ValuesListQuerySet(id1, id2, ...) ),
                (ModelClass_2, ValuesListQuerySet(id1, id2, id3 ...) )
                ...
                (ModelClass_N, ValuesListQuerySet(id1, ...))
            )
        """
        # Находим все соотношения с другими объектами:
        _relations = cls._meta.get_all_related_objects()
        result = []
        for rel in _relations:
            # Имя поля, которое ссылается на наш объект в зависимой модели:
            _attname = rel.field.attname
            if _attname.endswith('_id'):
                _attname = _attname.replace('_id', '')
            _attname = '%s__pk' % _attname
            # Пытаемся определить наличие зависимых объектов:
            pks = rel.field.model.objects.filter(
                **{str(_attname): obj_id}
            ).values_list('pk')
            if pks.exists():
                result.append((rel.field.model, pks))

        return tuple(result)

    @classmethod
    def get_verbose_object_relations(cls, obj_id):
        """
        Функция для определения наименований моделей,
        объектов зависимых от экземпляра данного класса.
        Экземпляр класса задается через cls и obj_id

        :param cls: Класс экземпляра, от которого будем искать зависимых
        :param obj_id: ID экземпляра от которого будем искать зависимых
        :return: Кортеж вида:
            (
                u"ModelClassName_1",
                u"ModelClassName_2",
                ...
                u"ModelClassName_N",
            )
        """
        relations = cls.get_object_relations(obj_id)
        return tuple([rel[0]._meta.verbose_name for rel in relations])

    class Meta:
        abstract = True


class BaseDictionaryModel(BaseModel):
    """Базовая модель справчоника."""

    code = models.CharField(
        max_length=100, null=True, blank=True, verbose_name='Код'
    )
    name = models.CharField(max_length=250, verbose_name='Наименование')

    class Meta:
        abstract = True
        unique_together = ('code', 'name')

    def save(self, *args, **kwargs):
        try:
            super().save(*args, **kwargs)
        except IntegrityError as err:
            if is_duplicate_integrity_error(str(err)):
                raise ItemAlreadyExistError
            else:
                raise err


class SimpleDictionaryModel(BaseDictionaryModel):
    """Модель простого справочника."""

    NON_DELETABLE_CODES = []
    NON_EDITABLE_CODES = []

    def __str__(self):
        return str(self.name) or ''

    def safe_delete(self):
        if self.code in self.NON_DELETABLE_CODES:
            raise ApplicationLogicException(
                'Элемент с этим кодом нельзя удалить. '
                'Он используется для автоматических операций!')

        return super(SimpleDictionaryModel, self).safe_delete()

    def save(self, *args, **kwargs):

        if self.code in self.NON_EDITABLE_CODES:
            if self.pk is not None:
                raise ValidationError(
                    'Элемент с этим кодом нельзя редактировать!')
            else:
                raise ValidationError(
                    'Элемент с указанным кодом имеется в системе. '
                    'Повторное создание невозможно.')

        super().save(*args, **kwargs)

    class Meta:
        abstract = True
        unique_together = ('code', 'name')


SystemDictionaryMeta = model_modifier_metaclass(**{
    "name": {
        "max_length": 100,
    }
})


class SystemDictionaryModel(
    ReadOnlyMixin,
    BaseDictionaryModel,
    metaclass=SystemDictionaryMeta
):
    """Справочник c системными значениями."""
    code = models.PositiveSmallIntegerField(
        verbose_name='Код',
    )
    is_system = models.BooleanField(
        verbose_name="Системная запись",
        default=False,
    )

    def is_read_only(self):
        """Возвращает признак доступа на изменение/удаление записи."""
        return self.is_system and self.pk

    @staticmethod
    def get_read_only_error_message(delete=False):
        action = "Удаление" if delete else "Изменение"
        return f"Выбранная запись является системной. {action} невозможно."

    class Meta:
        abstract = True
        unique_together = ('code', 'name')
